
/**
 * For more information see the head comment within the 'java.g' grammar file
 * that defines the input for this tree grammar.
 *
 * BSD licence
 * 
 * Copyright (c) 2007-2008 by HABELITZ Software Developments
 *
 * All rights reserved.
 * 
 * http://www.habelitz.com
 *
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. The name of the author may not be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY HABELITZ SOFTWARE DEVELOPMENTS ('HSD') ``AS IS'' 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL 'HSD' BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, 
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
tree grammar JavaTreeParser;

options {
    backtrack = true; 
    memoize = true;
    tokenVocab = Java;
    ASTLabelType = CommonTree;
}


@treeparser::header {
 package edu.cs.umu.sres.output;
 import edu.cs.umu.sres.SRES; 
 import edu.cs.umu.sres.Halstead;
}

@treeparser::members {
    
    // @Nadeem
    // to count unique operators, unique operands, total operators and total operands respectively, according to Halstead's measures respectively
    int n1=0;
    int n2=0;
    int N1=0;
    int N2=0; 
    
    int N=0; // Program length
    int n=0; //vocabulary
 
    List ODList = new ArrayList();
    List OPList = new ArrayList();
    
    //////////////////////////////////////////////////////////////
    //SRES Measures
    int syllablesCount=0;
    int wordsCount=0;    
    int wordsLength=0;    
    int sentencesCount=0;

// to count both operators, n1 and N1
    public void updateOPCounter(String op ){
	if(op!=null){
	 if(!OPList.contains(op)){
    		OPList.add(op);
    		n1++;
    		N1++;
    	}//inner if    		
    	else
    		N1++;
     }//outer if
    }
 
    // to count operators , op is operator and n is number of times operator is to be counted
    public void updateOPCounter(String op, int n ){
	if(op!=null){
	 if(!OPList.contains(op)){
    		OPList.add(op);
    		n1++;
    		N1+=n;
    	}//inner if    		
    	else
     		N1+=n;
     }//outer if
    }


 // to count operands, n2 and N2 
    public void updateODCounter(String a ){
	if(a!=null ){
	 if(!ODList.contains(a)){
    		ODList.add(a);
    		n2++;
    		N2++;
    	}//inner if    		
    	else
    		N2++;
     }//outer if	
    }
    
  // To count an operand based on the boolean flag , this method is used to count local varibles with same name but different scopes
    public void updateODCounter(String a, boolean f ){
     if(f){
	 ODList.add(a);
    	 n2++;
    	 N2++;
     }
     else{    
	if(a!=null ){
	 if(!ODList.contains(a)){
    		ODList.add(a);
    		n2++;
    		N2++;
    	}//inner if    		
    	else
    		N2++;
    	}//inner else	
     }//outer else
    }
       
  //To count method call as an operator    
    public void methodCallHandler(String idStr, String arg ){
  	if (idStr.contains(".")) {
            
            String parts[] = idStr.split("\\.");
            for(int i=0; i <parts.length; i++){
              ODList.remove(parts[i]);
	      n2--; N2--;
            }
            updateOPCounter(idStr+arg);
	}
	else{ // a method name without any class or object prefix
	   ODList.remove(idStr);
	   n2--; N2--;
	   updateOPCounter(idStr+arg);
    	}
    }
    
    
   //////////////////////////////////////////////////////////////
    public void addToWordsCount(int n){
     wordsCount+=n;
    }
    
    public void addToWordsLength(int n){
     wordsLength+=n;
    }
   
   /////////////////////////////////////////////////////////////// 
    //To set Length of the program
    public void setLength(int a){
     N=a;
    }
    
    //To get Length of the program
    public int getLength(){
     return N;
    }
    
        //To set Vocabulary of the program
    public void setVocabulary(int a){
     n=a;
    }
    
    //To get Vocabulary of the program
    public int getVocabulary(){
     return n;
    }
 
    
    ///////////////////////////////////////////////////////////////////////////////
    boolean mMessageCollectionEnabled = false;
    private boolean mHasErrors = false;
    List<String> mMessages;

    /**
     *  Switches error message collection on or of.
     *
     *  The standard destination for parser error messages is <code>System.err</code>.
     *  However, if <code>true</code> gets passed to this method this default
     *  behaviour will be switched off and all error messages will be collected
     *  instead of written to anywhere.
     *
     *  The default value is <code>false</code>.
     *
     *  @param pNewState  <code>true</code> if error messages should be collected.
     */
    public void enableErrorMessageCollection(boolean pNewState) {
        mMessageCollectionEnabled = pNewState;
        if (mMessages == null && mMessageCollectionEnabled) {
            mMessages = new ArrayList<String>();
        }
    }
    
    /**
     *  Collects an error message or passes the error message to <code>
     *  super.emitErrorMessage(...)</code>.
     *
     *  The actual behaviour depends on whether collecting error messages
     *  has been enabled or not.
     *
     *  @param pMessage  The error message.
     */
     @Override
    public void emitErrorMessage(String pMessage) {
        if (mMessageCollectionEnabled) {
            mMessages.add(pMessage);
        } else {
            super.emitErrorMessage(pMessage);
        }
    }
    
    /**
     *  Returns collected error messages.
     *
     *  @return  A list holding collected error messages or <code>null</code> if
     *           collecting error messages hasn't been enabled. Of course, this
     *           list may be empty if no error message has been emited.
     */
    public List<String> getMessages() {
        return mMessages;
    }
    
    /**
     *  Tells if parsing a Java source has caused any error messages.
     *
     *  @return  <code>true</code> if parsing a Java source has caused at least one error message.
     */
    public boolean hasErrors() {
        return mHasErrors;
    }
}

// Starting point for parsing a Java file.
javaSource
@init{

}
@after{ 
	setLength(N1+N2);
	setVocabulary(n1+n2);
	
	//to update SRES fields:
	SRES.setWordsCount(wordsCount);
	SRES.setWordsLength(wordsLength);
	SRES.setSentencesCount(sentencesCount);

	//to update Halstead's fields:
	Halstead.setDistinctOperators(n1);
	Halstead.setDistinctOperands(n2);
	Halstead.setTotalOperators(N1);
	Halstead.setTotalOperands(N2);
	Halstead.setProgramLength(N);
	Halstead.setProgramVocabulary(n);
	
	System.out.println("n1:No. of distinct operators = "+n1+
			"\nn2:No. of distinct operands = "+n2+
			"\nN1:Total No. of operators = "+N1+
			"\nN2:Total No. of operands = "+N2+
			"\n\nProgram Length = "+N+
			"\n\nProgram Vocabulary = "+n+			
			"\n\nOP List: "+OPList+"\nOD List: "+ODList+
			"\n\n\n----------SRES Measures--------- "+
			"\nSRES-Words = "+SRES.getWordsCount()+
			"\nSRES Words' Length = "+SRES.getWordsLength()+
			"\nSRES Sentences = "+SRES.getSentencesCount());
			
			}

    :   ^(JAVA_SOURCE annotationList packageDeclaration? importDeclaration* typeDeclaration*)
    ;

packageDeclaration
    :   ^(PACKAGE qualifiedIdentifier SEMI)  {addToWordsCount(1); addToWordsLength($PACKAGE.text.length()); sentencesCount++;}//SRES
    	{updateOPCounter("package"); updateOPCounter($SEMI.text); }

    ;
    
importDeclaration
    :   ^(IMPORT STATIC? qualifiedIdentifier DOTSTAR?)
    {addToWordsCount(1); addToWordsLength($IMPORT.text.length()); sentencesCount++;
     if($STATIC.text!=null){addToWordsCount(1); addToWordsLength(6);}
     if($DOTSTAR.text!=null){addToWordsCount(1); addToWordsLength(2);}
    }
    
    //Halstead's measures
    {updateOPCounter("import"); updateOPCounter(";");
     if($STATIC.text!=null){updateOPCounter("static");}
     if($DOTSTAR.text!=null){updateOPCounter(".*");}
    }
    
    ;
    
typeDeclaration
    :   ^(CLASS modifierList IDENT genericTypeParameterList? extendsClause? implementsClause? classTopLevelScope)
    	{  
    	 addToWordsCount(2); //2 words counted are "class" and IDENT here.
   	 addToWordsLength($IDENT.text.length()+ $CLASS.text.length());
   	}
   	{updateOPCounter("class"); updateODCounter($IDENT.text);}

    |   ^(INTERFACE modifierList IDENT genericTypeParameterList? extendsClause? interfaceTopLevelScope)
   	 {addToWordsCount(2); //2 words counted are "class" and IDENT here.
   	  addToWordsLength($IDENT.text.length()+ $INTERFACE.text.length());
   	 }
   	 {updateOPCounter("interface"); updateODCounter($IDENT.text);}
    |   ^(ENUM modifierList IDENT implementsClause? enumTopLevelScope)
    	{addToWordsCount(2); //2 words counted are "enum" and IDENT here.
   	 addToWordsLength($IDENT.text.length()+ 4); //4 for "enum 
   	}
   	{updateOPCounter("enum"); updateODCounter($IDENT.text);}
   	
    |   ^(AT modifierList IDENT annotationTopLevelScope)
   	{addToWordsCount(2); //2 words counted are "@interface" and IDENT here.
    	 addToWordsLength($IDENT.text.length()+ 10); //10 for "@interface"
    	}
     	{updateOPCounter("@interface"); updateODCounter($IDENT.text);}
    ;

extendsClause // actually 'type' for classes and 'type+' for interfaces, but this has 
              // been resolved by the parser grammar.
    :   ^(EXTENDS_CLAUSE typeList) {addToWordsCount(1); addToWordsLength(7); } // length of "extends" = 7, otherwise can be achieved by addToWordsLength(new String("extends").length());}
//        :   ^(EXTENDS_CLAUSE type+) {addToWordsCount(1); addToWordsLength(7); } // length of "extends" = 7, otherwise can be achieved by addToWordsLength(new String("extends").length());}
    	{updateOPCounter("extends");}
    ;   
    
implementsClause
    :   ^(IMPLEMENTS_CLAUSE typeList){addToWordsCount(1); addToWordsLength(10);} //length of "implements" = 10 otherwis, addToWordsLength(new String("implements").length());}
// :   ^(IMPLEMENTS_CLAUSE type+){addToWordsCount(1); addToWordsLength(10);} //length of "implements" = 10 otherwis, addToWordsLength(new String("implements").length());}
    	{updateOPCounter("implements");}
    ;

typeList
    :   type (commas +=COMMA type)*
    	{
    	 if($commas!=null){
    	  updateOPCounter(",", $commas.size());
    	  }
    	}
    ;  
//TODO count SRES for genericTypeParameters... as below        
genericTypeParameterList
    :   ^(GENERIC_TYPE_PARAM_LIST COMMA* genericTypeParameter+) //{System.out.println(" genericTypeParameter: "+$genericTypeParameter.text);}
    	{updateOPCounter("<>");
    	 //System.out.println(" generics comma: "+$COMMA.text);
    	 if($COMMA.text!=null)
    	  updateOPCounter(",");
    	} //generic operators
    ;

genericTypeParameter
    :   ^(IDENT bound?){addToWordsCount(1); addToWordsLength($IDENT.text.length());}//{System.out.println(" genericTypeParameter IDENT: "+$IDENT.text+" , bound: "+$bound.text);}
    	{updateODCounter($IDENT.text);}
    ;
        
bound
    :   ^(EXTENDS_BOUND_LIST type+)//{System.out.println(" genericTypeParameter bound: "+$type.text);}
    ;

enumTopLevelScope
      :   ^(ENUM_TOP_LEVEL_SCOPE enumConstant* SEMI? classTopLevelScope?)
//    :   ^(ENUM_TOP_LEVEL_SCOPE enumConstant+ COMMA? SEMI? classTopLevelScope?)
    	{
    	 if($SEMI.text!=null)
    	  updateOPCounter(";");
    	    	 
    	 if($classTopLevelScope.text==null){ 
    	  sentencesCount++; 
    	  updateOPCounter("{}");
    	 }    	  
    	 
    	 if($enumConstant.text!= null)
    	   sentencesCount++;
     	}
    ;

enumConstant
    :   ^(IDENT annotationList arguments? COMMA? classTopLevelScope?)
    	{
    	 if(!$IDENT.text.equals("<missing IDENT>")){    	 
       	   addToWordsCount(1); addToWordsLength($IDENT.text.length());
       	   updateODCounter($IDENT.text);
       	 }
       	 if($COMMA.text!=null)
    	  updateOPCounter(",");
       	  	 
       	}
    ;
    
    
classTopLevelScope 
    :   ^(CLASS_TOP_LEVEL_SCOPE classScopeDeclarations*){updateOPCounter("{}"); sentencesCount++;}
    ;

//TODO adjust for genericTypeParameterList    
classScopeDeclarations //These are class level declarations, and not method level declarations.
    :   ^(CLASS_INSTANCE_INITIALIZER block)//{sentencesCount++;}// updateOPCounter("{}");} //sentence is counted against a block { ... } in a class.
    |   ^(CLASS_STATIC_INITIALIZER block){addToWordsCount(1); addToWordsLength(6);}// sentencesCount++;}// word counted is "static" for static code block {System.out.println("CLASS_STATIC_INITIALIZER ");}
    	{ updateOPCounter("static");}
    	
    |   ^(FUNCTION_METHOD_DECL modifierList genericTypeParameterList? type IDENT formalParameterList arrayDeclaratorList? throwsClause? block?)
	{addToWordsCount(1); addToWordsLength($IDENT.text.length());}// sentencesCount++;}// word counted is IDENT"
	{updateODCounter($IDENT.text);  }
	
    |   ^(VOID_METHOD_DECL modifierList genericTypeParameterList? IDENT formalParameterList throwsClause? block?) 
	{addToWordsCount(2); addToWordsLength($IDENT.text.length()+4);}//4 for "void, here the 2 words counted are "void" and IDENT"
	{updateODCounter($IDENT.text); updateOPCounter("void");  }
	
    |   ^(VAR_DECLARATION modifierList type variableDeclaratorList SEMI) {sentencesCount++;} // as every variable declaration is a 1 sentence.
    	{updateOPCounter($SEMI.text);}
    	
    |   ^(CONSTRUCTOR_DECL modifierList genericTypeParameterList? IDENT formalParameterList throwsClause? block)
    	{addToWordsCount(1); addToWordsLength($IDENT.text.length());}
    	{updateODCounter($IDENT.text); }
    	
    |   typeDeclaration
    |	SEMI
        {updateOPCounter($SEMI.text); sentencesCount++;}
    ;
    
interfaceTopLevelScope
    :   ^(INTERFACE_TOP_LEVEL_SCOPE interfaceScopeDeclarations*){sentencesCount++; updateOPCounter("{}");}
     ;
    
interfaceScopeDeclarations
    :   ^(FUNCTION_METHOD_DECL modifierList genericTypeParameterList? type IDENT formalParameterList arrayDeclaratorList? throwsClause?)
    	{addToWordsCount(1); addToWordsLength($IDENT.text.length()); sentencesCount++;}// word counted is IDENT"
    	{updateODCounter($IDENT.text); updateOPCounter(";"); }
    |   ^(VOID_METHOD_DECL modifierList genericTypeParameterList? IDENT formalParameterList throwsClause?)
	{addToWordsCount(2); addToWordsLength($IDENT.text.length()+4);sentencesCount++;}//4 for "void, here the 2 words counted are "void" and IDENT"
	{updateODCounter($IDENT.text); updateOPCounter("void"); updateOPCounter(";"); }
                         // Interface constant declarations have been switched to variable
                         // declarations by 'java.g'; the parser has already checked that
                         // there's an obligatory initializer.
    |   ^(VAR_DECLARATION modifierList type variableDeclaratorList){sentencesCount++;} // as every variable declaration is a 1 sentence.
    	{updateOPCounter(";");}
    	
    |   typeDeclaration
    |   SEMI {updateOPCounter($SEMI.text); sentencesCount++;}
    ;

variableDeclaratorList
:   ^(VAR_DECLARATOR_LIST variableDeclarator (commas+=COMMA variableDeclarator)*)
	{
    	 if($commas!=null){
    	  updateOPCounter(",", $commas.size());
    	  }
    	}
//    :   ^(VAR_DECLARATOR_LIST variableDeclarator+)
    ;

variableDeclarator
    :   ^(VAR_DECLARATOR variableDeclaratorId b=variableInitializer?)	
        {    	    	
    	 if($b.text!=null){
 	     updateOPCounter("=");
 	     addToWordsCount(1); addToWordsLength(1); //for SRES measure, we count "=" (operator) as a word and lenghth is 1 in case of "="
 	    // System.out.println("= is counted as word");
 	 }
     	}
    ;
    
variableDeclaratorId
    :   ^(IDENT arrayDeclaratorList?) {addToWordsCount(1); addToWordsLength($IDENT.text.length());}//{System.out.println("IDENT.text "+$IDENT.text);}
    	{updateODCounter($IDENT.text, true);}
    ;

variableInitializer
    :   arrayInitializer //{addToWordsCount(1); addToWordsLength(1); System.out.println("Array Initialzier found "+$arrayInitializer.text);}//{ }// we count "=" (every operator) as a word and lenghth is 1 in case of "="
    |   expression //{ addToWordsCount(1); addToWordsLength(1);} // we count "=" (every operator) as a word and lenghth is 1 in case of "="
    ;

arrayDeclarator
    //:   LBRACK RBRACK {addToWordsCount(1); addToWordsLength(2); System.out.println("Array Declaration found ");} //2 for "[ ]"
    :	ARRAY_DECLARATOR {addToWordsCount(1); addToWordsLength(2); }//System.out.println("Array Declaration found ");} //2 for "[ ]"
     {updateOPCounter("[]"); } //array operator
    ;

arrayDeclaratorList
    //:   ^(ARRAY_DECLARATOR_LIST ARRAY_DECLARATOR*)  {addToWordsCount(1); addToWordsLength(2); System.out.println("Array Declaration found ");} //2 for "[ ]"
    :   ^(ARRAY_DECLARATOR_LIST arrayDeclarator+)  
    ;
    
arrayInitializer
//    :   ^(ARRAY_INITIALIZER (variableInitializer (commas+=COMMA variableInitializer)* c=COMMA?)?)
    :   ^(ARRAY_INITIALIZER (variableInitializer commas+=COMMA* variableInitializer* )?)
    	{
    	 if($commas!=null){
    	  //System.out.println("Commas are "+$commas.size());
    	  updateOPCounter(",", $commas.size());
    	  }
    	 
    	  updateOPCounter("{}"); 
    	}
    ;

throwsClause
    :   ^(THROWS_CLAUSE qualifiedIdentifier (commas+=COMMA qualifiedIdentifier)*) 
    	{addToWordsCount(1); addToWordsLength(6);} // 6 For "throws"
    	
    	{
    	 if($commas!=null){
    	  updateOPCounter(",", $commas.size());
    	  }
    	 updateOPCounter("throws"); 
    	}
    ;

modifierList
    :   ^(MODIFIER_LIST modifier*)
    ;

modifier
    :   PUBLIC {wordsCount++; addToWordsLength($PUBLIC.text.length());}  {updateOPCounter("public"); }
    |   PROTECTED {wordsCount++; addToWordsLength($PROTECTED.text.length());} {updateOPCounter("protected"); }
    |   PRIVATE {wordsCount++; addToWordsLength($PRIVATE.text.length()); } {updateOPCounter("private"); }
    |   STATIC {wordsCount++; addToWordsLength($STATIC.text.length());}{updateOPCounter("static"); }
    |   ABSTRACT {wordsCount++; addToWordsLength($ABSTRACT.text.length());}{updateOPCounter("abstract"); }
    |   NATIVE {wordsCount++; addToWordsLength($NATIVE.text.length());}{updateOPCounter("native"); }
    |   SYNCHRONIZED {wordsCount++; addToWordsLength($SYNCHRONIZED.text.length());}{updateOPCounter("synchrblockonized"); }
    |   TRANSIENT {wordsCount++; addToWordsLength($TRANSIENT.text.length());}{updateOPCounter("transient"); }
    |   VOLATILE {wordsCount++; addToWordsLength($VOLATILE.text.length());}{updateOPCounter("volatile"); }
    |   STRICTFP {wordsCount++; addToWordsLength($STRICTFP.text.length());}{updateOPCounter("strictfp"); }
    |   localModifier
    ;

localModifierList
    :   ^(LOCAL_MODIFIER_LIST localModifier*)
    ;

localModifier
    :   FINAL {wordsCount++; addToWordsLength($FINAL.text.length());} {updateOPCounter("final"); }    
    |   annotation
    ;

type
    :   ^(TYPE (a=primitiveType | qualifiedTypeIdent) arrayDeclaratorList? )
    	{
    	   updateOPCounter($a.text); //for primitive types  
    	   //if($COMMA.text!=null)
    	  //	updateOPCounter(",");	  
   	}
    ;

qualifiedTypeIdent
    :   ^(QUALIFIED_TYPE_IDENT typeIdent+) 
    ;

typeIdent
    :   ^(IDENT genericTypeArgumentList?) {addToWordsCount(1); addToWordsLength($IDENT.text.length());}
    	{ 
    	// updateIdODCounter($IDENT.text); 
    	 updateOPCounter($IDENT.text); // for user defined types
    	}
    ;

primitiveType
    :   BOOLEAN {addToWordsCount(1); addToWordsLength($BOOLEAN.text.length());}
    |   CHAR    {addToWordsCount(1); addToWordsLength($CHAR.text.length());}
    |   BYTE    {addToWordsCount(1); addToWordsLength($BYTE.text.length());}
    |   SHORT    {addToWordsCount(1); addToWordsLength($SHORT.text.length());}
    |   INT    {addToWordsCount(1); addToWordsLength($INT.text.length());}
    |   LONG    {addToWordsCount(1); addToWordsLength($LONG.text.length());}
    |   FLOAT    {addToWordsCount(1); addToWordsLength($FLOAT.text.length());}
    |   DOUBLE    {addToWordsCount(1); addToWordsLength($DOUBLE.text.length());}
    ;

genericTypeArgumentList
    :   ^(GENERIC_TYPE_ARG_LIST genericTypeArgument (commas+=COMMA genericTypeArgument)*) //{System.out.println("genericTypeArgument type: "+$genericTypeArgument.text);}
      	{
      	 if($commas!=null){
    	  updateOPCounter(",", $commas.size());
    	  }
      	 updateOPCounter("<>");
      	} //generic operators
    ;
    
genericTypeArgument
    :   type 
    |   ^(QUESTION genericWildcardBoundType?){addToWordsCount(1); addToWordsLength(1);}//word counted is "?" a wild card in generics
    	{updateOPCounter("?:gen");} //? generic wildcard ef used to ref a supertype.
    ;

genericWildcardBoundType                                                                                                                      
    :   ^(EXTENDS type){addToWordsCount(1); addToWordsLength(7);}// word counted is "extends"
    	{updateOPCounter("extends");} 
    |   ^(SUPER type){addToWordsCount(1); addToWordsLength(5);}// word counted is "super"
    	{updateOPCounter("super");} 
    ;

formalParameterList
//    :   ^(FORMAL_PARAM_LIST formalParameterStandardDecl* formalParameterVarargDecl?) 
    :   ^(FORMAL_PARAM_LIST formalParameterStandardDecl? commas+=COMMA* formalParameterStandardDecl* formalParameterVarargDecl? ) 
    	{
    	 if($commas!=null){
    	  updateOPCounter(",", $commas.size());
    	  }
    	 
    	}
    
    //|	^(FORMAL_PARAM_LIST formalParameterVarArgDecl) 
    ;
    
formalParameterStandardDecl
    :   ^(FORMAL_PARAM_STD_DECL localModifierList type variableDeclaratorId)
    ;
    
formalParameterVarargDecl
    :   ^(FORMAL_PARAM_VARARG_DECL localModifierList type variableDeclaratorId){System.out.println("\nWANTED---formalParameterVarargDecl\n");}
    ;
    
qualifiedIdentifier
    :   IDENT    {addToWordsCount(1); addToWordsLength($IDENT.text.length()); }//System.out.println("-...--Its me doing bad");}
    {updateODCounter($IDENT.text);}
    |   ^(DOT qualifiedIdentifier IDENT) {addToWordsCount(1+1); addToWordsLength($IDENT.text.length()+1); }// One is added for dot operator}
    {updateODCounter($IDENT.text); updateOPCounter(".");}
//    |   ^(DOT qualifiedIdentifier IDENT) {addToWordsCount(1*2); addToWordsLength($IDENT.text.length()*2); }//System.out.println("---Its me doing bad");}
    ;
    
// ANNOTATIONS

annotationList
    :   ^(ANNOTATION_LIST annotation*)
    ;

annotation
    :   ^(AT qualifiedIdentifier annotationInit?) //{System.out.println("Identifier : "+$qualifiedIdentifier.text);}
    	{ addToWordsLength(1); sentencesCount++;} //1 for @ , {addToWordsCount(1);} words counted is "@"
    	{ updateOPCounter("@"); }
    	
    ;
    
annotationInit
    :   ^(ANNOTATION_INIT_BLOCK annotationInitializers)//{System.out.println(" annotationInitializers: "+$annotationInitializers.text);}
    ;

annotationInitializers
//    :   ^(ANNOTATION_INIT_KEY_LIST annotationInitializer+)//{System.out.println(" annotationInitializer: "+$annotationInitializer.text);}
    :   ^(ANNOTATION_INIT_KEY_LIST annotationInitializer (commas+=COMMA annotationInitializer)*)//{System.out.println(" annotationInitializer: "+$annotationInitializer.text);}
    	{
    	 if($commas!=null){
    	  updateOPCounter(",", $commas.size());
    	  }
    	}
    |   ^(ANNOTATION_INIT_DEFAULT_KEY annotationElementValue)//{System.out.println("annotationElementValue : "+$annotationElementValue.text);}
    ;
    
    
annotationInitializer
    :   ^(IDENT annotationElementValue){addToWordsCount(2); addToWordsLength($IDENT.text.length()+1);}// 2 words counted are IDENT and "=" operator {System.out.println(" annotationInitializer IDENT: "+$IDENT.text);}
    	 {updateODCounter($IDENT.text); updateOPCounter("="); }

    ;
    
annotationElementValue
//    :   ^(ANNOTATION_INIT_ARRAY_ELEMENT annotationElementValue*)
    :	^(ANNOTATION_INIT_ARRAY_ELEMENT (annotationElementValue commas+=COMMA* annotationElementValue*) ?)
    {
    	 if($commas!=null){
    	  //System.out.println("Commas are "+$commas.size());
    	  updateOPCounter(",", $commas.size());
    	  }
    	 
    	  updateOPCounter("{}"); 
    	}
    |   annotation
    |   expression //{System.out.println(" annotationInitializer annotationElementValue: "+$expression.text);}
    ;
    
annotationTopLevelScope
    :   ^(ANNOTATION_TOP_LEVEL_SCOPE annotationScopeDeclarations*){sentencesCount++;}
    	 { updateOPCounter("{}"); }
    ;
    
annotationScopeDeclarations
    :   ^(ANNOTATION_METHOD_DECL modifierList type IDENT annotationDefaultValue?)
	{addToWordsCount(1); addToWordsLength($IDENT.text.length()); sentencesCount++;}// word counted is IDENT"
    	{updateODCounter($IDENT.text); updateOPCounter(";"); }
    |   ^(VAR_DECLARATION modifierList type variableDeclaratorList){sentencesCount++;} // as every variable declaration is a 1 sentence.
    	{updateOPCounter(";"); }
    |   typeDeclaration
    ;
    
annotationDefaultValue
    :   ^(DEFAULT annotationElementValue) {addToWordsCount(1); addToWordsLength($DEFAULT.text.length()); }// word counted is "default"
 	{updateOPCounter("defualt:ann"); }   // :ann is added to distinguish default keyword used under switch statement. 
    ;

// STATEMENTS / BLOCKS

block
    :   ^(BLOCK_SCOPE blockStatement*){updateOPCounter("{}"); sentencesCount++;}
    ;
    
blockStatement
    :   localVariableDeclaration 
    |   typeDeclaration
    |   statement //{System.out.println("\n statement: "+$statement.text);}
    	
    ;
    
localVariableDeclaration
    :   ^(VAR_DECLARATION localModifierList type variableDeclaratorList) {sentencesCount++; }// every local declaration is a sentence
    	{updateOPCounter(";");}
    ;
    

statement
    :   block
    |	^(ASSERT expression COLON? expression? SEMI)
        {//System.out.println("assert Colon is "+$COLON.text);
         sentencesCount++; addToWordsCount(1); addToWordsLength(6); //words counted is "assert"
         updateOPCounter("assert"); updateOPCounter(";");
    	 if($COLON.text!=null){
    	  addToWordsCount(1); addToWordsLength(1); //words counted here is ":" in case of alternative assert Expression1 : Expression2 
    	  updateOPCounter(":assert");
    	 }
    	} 
	
//TODO check for else part of if.	
    |   ^(IF parenthesizedExpression a=statement elseStatement?) {addToWordsCount(1); addToWordsLength(2);}//words counted is "if"
	{updateOPCounter("if"); }  //{System.out.println("if stat: "+$a.text);}
//TODO problem with for due to block statement.	 also check sentencesCount--; in forInit rule
    |   ^(FOR forInit SEMI forCondition SEMI forUpdater statement){addToWordsCount(1); addToWordsLength(3);}//words counted is "for"
    	{updateOPCounter("for");  updateOPCounter(";");}  //; counted for forCondition, ; for forInit is counted in its rule  	 
    |   ^(FOR_EACH localModifierList type IDENT expression statement) 
        {addToWordsCount(3); addToWordsLength($IDENT.text.length()+3+1); } //3 words counted are "IDENT", "for" and ":" respectively
        {updateOPCounter("for");  updateOPCounter(":(for each)"); updateODCounter($IDENT.text);}
            	 
    |   ^(WHILE parenthesizedExpression statement){addToWordsCount(1); addToWordsLength(5);}//{System.out.println("\n while stat: "+$a.text);}  //word counted is "while"
    	 {updateOPCounter("while"); }
    	 
    |   ^(DO statement WHILE parenthesizedExpression SEMI){addToWordsCount(1); addToWordsLength(2);}//words counted is "do"
         {updateOPCounter("do"); updateOPCounter("while");updateOPCounter(";");}
    
    |   ^(TRY block catches? b=block?) 
    	 {
    	   addToWordsCount(1); addToWordsLength(3); //words counted is "try"
    	   updateOPCounter("try");
    	   if($b.text!=null){ // The b=block is the optional finally block.
           addToWordsCount(1); addToWordsLength(7); //words counted is "finally"
           updateOPCounter("finally");
           //sentencesCount++;
           }
          }  
    |   ^(SWITCH parenthesizedExpression switchBlockLabels){addToWordsCount(1); addToWordsLength(6); sentencesCount++;}//6 for "switch"
    {updateOPCounter("switch");}//System.out.println("----Switch stat - switchBlockLabels: "+$switchBlockLabels.text);}
    
    |   ^(SYNCHRONIZED parenthesizedExpression block){addToWordsCount(1); addToWordsLength(12);}//12 for "synchronized"
	{updateOPCounter("synchronized");}
    
    |   ^(RETURN expression? SEMI) {addToWordsCount(1); addToWordsLength(6); sentencesCount++;}//6 for "return"
    	{updateOPCounter("return");updateOPCounter(";");}
    	
    |   ^(THROW expression SEMI){addToWordsCount(1); addToWordsLength(5); sentencesCount++;}//5 for "throw"
    	{updateOPCounter("throw"); updateOPCounter(";");}
    	
    |   ^(BREAK IDENT? SEMI){addToWordsCount(1); addToWordsLength(5); sentencesCount++;//word counted is "break"
    	updateOPCounter("break"); updateOPCounter(";");
     	  if($IDENT.text!=null){addToWordsCount(1); addToWordsLength($IDENT.text.length()); updateODCounter("$IDENT.text"); }
         }
         
    |   ^(CONTINUE IDENT? SEMI){addToWordsCount(1); addToWordsLength(8); sentencesCount++;//word counted is "continue"
    	updateOPCounter("continue"); updateOPCounter(";");
     	  if($IDENT.text!=null){addToWordsCount(1); addToWordsLength($IDENT.text.length()); updateODCounter("$IDENT.text");}
         }
    |   ^(LABELED_STATEMENT IDENT statement){addToWordsCount(1); addToWordsLength($IDENT.text.length()+1); sentencesCount++;}//1 for ":" after label
    	{updateODCounter($IDENT.text); updateOPCounter(":label");}
    	
    |   expression SEMI {sentencesCount++; updateOPCounter(";"); } //System.out.println("\n sentence counted due to Expression: "+$expression.text); }//Every Expression corresponds to 1 sentence
    |   SEMI {addToWordsCount(1); addToWordsLength(1); sentencesCount++;}// Empty statement i.e only a ;. we count it one word of one lenght and a sentence.
    	{updateOPCounter(";");}
    ;

elseStatement
    :	^(ELSE_STATEMENT statement?){ updateOPCounter("else");}
    	{addToWordsCount(1); addToWordsLength(4);} // 4 for else
    ;     
       
catches
    :   ^(CATCH_CLAUSE_LIST catchClause+)
    ;
    
catchClause
    :   ^(CATCH formalParameterStandardDecl block){addToWordsCount(1); addToWordsLength(5); updateOPCounter("catch");}//words counted is "catch"
    ;

switchBlockLabels
    :   ^(SWITCH_BLOCK_LABEL_LIST switchCaseLabel* switchDefaultLabel? afterDefaultCaseLabel*)//TODO change name of last switchCaseLabel in both grammars
   
    ;
        
switchCaseLabel
    :   ^(CASE expression blockStatement*)
    {addToWordsCount(1); addToWordsLength(4+1); sentencesCount++;//words counted is "case" 1 is added for ":"
     updateOPCounter("case"); {updateOPCounter(":switch");}
//     sentencesCount--; // to nullify already counted sentence in blockStatement.
    }
    //{ System.out.println("----switchCaseLabel blockStatement: "+$blockStatement.text+"\n");}
    ;

afterDefaultCaseLabel
    :   ^(CASE expression blockStatement*)
    {addToWordsCount(1); addToWordsLength(4+1); sentencesCount++;//words counted is "case",  1 is added for ":" 
      updateOPCounter("case"); {updateOPCounter(":switch");}
//     sentencesCount--; // to nullify already counted sentence in blockStatement.
    }
    //{ System.out.println("----AfterDefaultCaseLabel blockStatement: "+$blockStatement.text+"\n");}
    ;
    
switchDefaultLabel
    :   ^(DEFAULT blockStatement*){addToWordsCount(1); addToWordsLength(7+1); sentencesCount++; //  1 is added for ":" 
      updateOPCounter("default:switch"); {updateOPCounter(":switch");}
    		//sentencesCount++; 
         }//words counted is "default"
    ;
    
forInit
    :   ^(FOR_INIT localVariableDeclaration) {sentencesCount--;} //to nullify the affect of sentence counted because of localVariableDeclaration rule
    |   ^(FOR_INIT expressionList) {updateOPCounter(";");}
    |   ^(FOR_INIT SEMI) {updateOPCounter(";");}
    ;
    
forCondition
    :   ^(FOR_CONDITION expression?)
    ;
    
forUpdater
//    :   ^(FOR_UPDATE expression*)
    :   ^(FOR_UPDATE expressionList?)
    ;
    
// EXPRESSIONS

parenthesizedExpression
    :   ^(PARENTESIZED_EXPR expression){updateOPCounter("()");}
    ;

expressionList
    :   expression (commas+=COMMA expression)*
    	{
    	 if($commas!=null){
    	  updateOPCounter(",", $commas.size());
    	  }
    	}
    ;
 
    
expression
    :   ^(EXPR expr)//{System.out.println("\n Expression: "+$expr.text);}
    ;

//every operator in expression corresponds to one word.
expr
    :   ^(op=ASSIGN expr expr){updateOPCounter($op.text); addToWordsCount(1);addToWordsLength(1);}// word counted is "=";}
    |   ^(op=PLUS_ASSIGN expr expr){updateOPCounter($op.text); addToWordsCount(1);addToWordsLength(2);}
    |   ^(op=MINUS_ASSIGN expr expr){updateOPCounter($op.text); addToWordsCount(1);addToWordsLength(2);}
    |   ^(op=STAR_ASSIGN expr expr){updateOPCounter($op.text); addToWordsCount(1);addToWordsLength(2);}
    |   ^(op=DIV_ASSIGN expr expr){updateOPCounter($op.text); addToWordsCount(1);addToWordsLength(2);}
    |   ^(op=AND_ASSIGN expr expr){updateOPCounter($op.text); addToWordsCount(1);addToWordsLength(2);}
    |   ^(op=OR_ASSIGN expr expr){updateOPCounter($op.text); addToWordsCount(1);addToWordsLength(2);}
    |   ^(op=XOR_ASSIGN expr expr){updateOPCounter($op.text); addToWordsCount(1);addToWordsLength(2);}
    |   ^(op=MOD_ASSIGN expr expr){updateOPCounter($op.text); addToWordsCount(1);addToWordsLength(2);}
    |   ^(op=BIT_SHIFT_RIGHT_ASSIGN expr expr){updateOPCounter($op.text); addToWordsCount(1);addToWordsLength(4);}//4 because of ">>>="
    |   ^(op=SHIFT_RIGHT_ASSIGN expr expr){updateOPCounter($op.text); addToWordsCount(1);addToWordsLength(3);}// ">>="
    |   ^(op=SHIFT_LEFT_ASSIGN expr expr){updateOPCounter($op.text); addToWordsCount(1);addToWordsLength(3);}// "<<="
    //TODO check length for QUESTION operator
    |   ^(op=QUESTION a=expr b=expr c=expr){updateOPCounter($op.text); addToWordsCount(1);addToWordsLength(1);} //{System.out.println("op: "+$op.text+", a: "+$a.text+", b: "+$b.text+", c: "+$c.text);}
    |   ^(op=LOGICAL_OR expr expr){updateOPCounter($op.text); addToWordsCount(1);addToWordsLength(2);}
    |   ^(op=LOGICAL_AND a=expr b=expr){updateOPCounter($op.text); addToWordsCount(1);addToWordsLength(2);} //{System.out.println("op: "+$op.text+", a: "+$a.text+", b: "+$b.text);}
    |   ^(op=OR a=expr b=expr){updateOPCounter($op.text);  addToWordsCount(1);addToWordsLength(1);}//{System.out.println("op: "+$op.text+", a: "+$a.text+", b: "+$b.text);}
    |   ^(op=XOR expr expr){updateOPCounter($op.text); addToWordsCount(1);addToWordsLength(1);}
    |   ^(op=AND expr expr){updateOPCounter($op.text); addToWordsCount(1);addToWordsLength(1);}
    |   ^(op=EQUAL expr expr){updateOPCounter($op.text); addToWordsCount(1);addToWordsLength(2);} //"=="
    |   ^(op=NOT_EQUAL expr expr){updateOPCounter($op.text); addToWordsCount(1);addToWordsLength(2);}
    |   ^(op=INSTANCEOF a=expr b=type) {updateOPCounter($op.text); addToWordsCount(1); addToWordsLength($INSTANCEOF.text.length());} //{System.out.println("op: "+$op.text+", a: "+$a.text+", b: "+$b.text);}
    |   ^(op=LESS_OR_EQUAL expr expr){updateOPCounter($op.text); addToWordsCount(1);addToWordsLength(2);}
    |   ^(op=GREATER_OR_EQUAL expr expr){updateOPCounter($op.text); addToWordsCount(1);addToWordsLength(2);}
    |   ^(op=BIT_SHIFT_RIGHT expr expr){updateOPCounter($op.text); addToWordsCount(1);addToWordsLength(3);}// ">>>"
    |   ^(op=SHIFT_RIGHT expr expr){updateOPCounter($op.text); addToWordsCount(1);addToWordsLength(2);}//">>"
    |   ^(op=GREATER_THAN a=expr b=expr){updateOPCounter($op.text); addToWordsCount(1);addToWordsLength(1);}//{System.out.println("op: "+$op.text+", a: "+$a.text+", b: "+$b.text);}
    |   ^(op=SHIFT_LEFT expr expr){updateOPCounter($op.text); addToWordsCount(1);addToWordsLength(2);}//"<<"
    |   ^(op=LESS_THAN expr expr){updateOPCounter($op.text); addToWordsCount(1);addToWordsLength(1);}
    |   ^(op=PLUS expr expr) {updateOPCounter($op.text); addToWordsCount(1);addToWordsLength(1);}
    |   ^(op=MINUS expr expr){updateOPCounter($op.text); addToWordsCount(1);addToWordsLength(1);}
    |   ^(op=STAR expr expr){updateOPCounter($op.text); addToWordsCount(1);addToWordsLength(1);}
    |   ^(op=DIV expr expr){updateOPCounter($op.text); addToWordsCount(1);addToWordsLength(1);}
    |   ^(op=MOD expr expr){updateOPCounter($op.text); addToWordsCount(1);addToWordsLength(1);}
    |   ^(op=UNARY_PLUS expr){updateOPCounter("UNARY_PLUS"); addToWordsCount(1);addToWordsLength(1);}
    |   ^(op=UNARY_MINUS expr) {updateOPCounter("UNARY_MINUS"); addToWordsCount(1);addToWordsLength(1);}
    |   ^(op=PRE_INC expr){updateOPCounter("PRE_INC"); addToWordsCount(1);addToWordsLength(2);}
    |   ^(op=PRE_DEC expr){updateOPCounter("PRE_DEC"); addToWordsCount(1);addToWordsLength(2);}
    |   ^(op=POST_INC expr){updateOPCounter("POST_INC"); addToWordsCount(1);addToWordsLength(2);}
    |   ^(op=POST_DEC expr){updateOPCounter("POST_DEC"); addToWordsCount(1);addToWordsLength(2);}
    |   ^(op=NOT expr) {updateOPCounter($op.text); addToWordsCount(1);addToWordsLength(1);}
    |   ^(op=LOGICAL_NOT expr){updateOPCounter($op.text); addToWordsCount(1);addToWordsLength(1);}
    //TODO count SRES for CAST EXPR
    |   ^(CAST_EXPR type expr) {updateOPCounter("()");}
    //for method call
    //|	^(METHOD_CALL a=primaryExpression genericTypeArgumentList? arguments){System.out.println(" METHOD Call primaryExpression: "+$a.text);} // To recognize method calls expression
    |   primaryExpression //{System.out.println("Primary Expression: "+$primaryExpression.text);}
    ;
    
primaryExpression
    :   ^(  DOT
            (   a=primaryExpression
                (   IDENT {addToWordsCount(1*2); addToWordsLength($IDENT.text.length()*2);}//{System.out.println("---Inner IDENT found: "+$IDENT.text);}
                    {updateODCounter($IDENT.text);}	
                |   THIS{addToWordsCount(1*2); addToWordsLength($THIS.text.length()*2);}//{System.out.println("---Inner This found");}
                    {updateOPCounter("this");}
                |   SUPER{addToWordsCount(1*2); addToWordsLength($SUPER.text.length()*2);}
                    {updateOPCounter("super");}
                |   innerNewExpression {System.out.println("\nWANTED---innerNewExpression Found in primaryExpression\n");}
                    {updateOPCounter("new");}
                |   CLASS {System.out.println("\nWANTED---CLASS Found in primaryExpression\n");}
                )
            |   primitiveType CLASS
            |   VOID CLASS
            ){addToWordsCount(1*2);addToWordsLength(1*2);} // "." is bit difficult to resolve in expressions, therefore we multiply it by factor x=2, and length aslo doubled //{System.out.println(" DOT : "+ $DOT.text);} // To recognize 
             {updateOPCounter(".");}
        )//{System.out.println(" DOT alternative: "+ $a.text);} // To recognize 
        
    |   parenthesizedExpression //{System.out.println(" parenthesizedExpression: "+ $parenthesizedExpression.text);}// To recognize parenthesized Expression e.g ((5+9)-6)*9; 
    |   IDENT {updateODCounter($IDENT.text); addToWordsCount(1); addToWordsLength($IDENT.text.length());}//{System.out.println(" outer IDENT i.e.operand of the expression: "+ $IDENT.text);} // To recognize (identifier) in the expression such as variable or method names
    //TODO: method call Hastead operator, count method call as operator.
    // write a method handler and pass it primaryExpression.text as argument and then updateValues accordingly.
    |   ^(METHOD_CALL a=primaryExpression genericTypeArgumentList? arguments)//{System.out.println(" METHOD Call primaryExpression: "+$a.text);} // To recognize method calls expression
           {methodCallHandler($a.text, $arguments.text);}
//    |   ^(METHOD_CALL a=IDENT b=DOT* c=IDENT* genericTypeArgumentList? arguments){System.out.println(" METHOD Call : "+$a.text+" b: "+$b.text+" c: "+$c.text);} // To recognize method calls expression
    |   explicitConstructorCall {System.out.println("\nWANTED---explicitConstructorCall Found in primaryExpression\n");}
    |   ^(ARRAY_ELEMENT_ACCESS a=primaryExpression b=expression) //{System.out.println("Array Element Access: "+ $a.text+" "+ $b.text);} // To recognize x[i]
    |   literal {updateODCounter($literal.text); addToWordsCount(1);addToWordsLength($literal.text.length());}//{System.out.println(" \nliteral of the expression: "+ $literal.text);}
    //TODO Check alternative in newExpression rule, for SRES measures.
    |   newExpression {addToWordsCount(1);addToWordsLength(3);}// word counted is "new" //to recognize new objects creation and array declaration
    	{updateOPCounter("new");}
    |   THIS {addToWordsCount(1);addToWordsLength(4);}/* 4 is length of "this"*///{System.out.println("---This found");}
    	{updateOPCounter("this");}
    |   arrayTypeDeclarator {System.out.println("\nWANTED---ArrayTypeDeclarator Found in primaryExpression\n");}
    |   SUPER{addToWordsCount(1);addToWordsLength(5);}
	{updateOPCounter("super");}
    ;
/*
methodCallExpression
 :   ^(  DOT
            (   a=methodCallExpression
                (   IDENT {addToWordsCount(1*2); addToWordsLength($IDENT.text.length()*2);}{System.out.println("---Inner IDENT found: "+$IDENT.text);}
                    {updateOPCounter($IDENT.text);}	
                |   THIS{addToWordsCount(1*2); addToWordsLength($THIS.text.length()*2);}//{System.out.println("---Inner This found");}
                    {updateOPCounter("this");}
                |   SUPER{addToWordsCount(1*2); addToWordsLength($SUPER.text.length()*2);}
                    {updateOPCounter("super");}
                |   innerNewExpression {System.out.println("\nWANTED---innerNewExpression Found in primaryExpression\n");}
                    {updateOPCounter("new");}
                |   CLASS {System.out.println("\nWANTED---CLASS Found in primaryExpression\n");}
                )
            |   primitiveType CLASS
            |   VOID CLASS
            ){addToWordsCount(1*2);addToWordsLength(1*2);} // "." is bit difficult to resolve in expressions, therefore we multiply it by factor x=2, and length aslo doubled //{System.out.println(" DOT : "+ $DOT.text);} // To recognize 
             {updateOPCounter(".");}
        )//{System.out.println(" DOT alternative: "+ $a.text);} // To recognize 
        
    |   parenthesizedExpression {updateOPCounter("()");}//{System.out.println(" parenthesizedExpression: "+ $parenthesizedExpression.text);}// To recognize parenthesized Expression e.g ((5+9)-6)*9; 
    |   IDENT {updateODCounter($IDENT.text); addToWordsCount(1); addToWordsLength($IDENT.text.length());}//{System.out.println(" outer IDENT i.e.operand of the expression: "+ $IDENT.text);} // To recognize (identifier) in the expression such as variable or method names
    //TODO: method call Hastead operator
    |   ^(METHOD_CALL methodCallExpression genericTypeArgumentList? arguments){System.out.println(" METHOD Call primaryExpression: "+$a.text);} // To recognize method calls expression
    |   explicitConstructorCall {System.out.println("\nWANTED---explicitConstructorCall Found in primaryExpression\n");}
    |   ^(ARRAY_ELEMENT_ACCESS a=methodCallExpression b=expression) //{System.out.println("Array Element Access: "+ $a.text+" "+ $b.text);} // To recognize x[i]
    |   literal {updateOPCounter($literal.text); addToWordsCount(1);addToWordsLength($literal.text.length());}//{System.out.println(" \nliteral of the expression: "+ $literal.text);}
    //TODO Check alternative in newExpression rule, for SRES measures.
    |   newExpression {addToWordsCount(1);addToWordsLength(3);}// word counted is "new" //to recognize new objects creation and array declaration
    	{updateOPCounter("new");}
    |   THIS {addToWordsCount(1);addToWordsLength(4);}// 4 is length of "this"{System.out.println("---This found");}
    	{updateOPCounter("this");}
    |   arrayTypeDeclarator {System.out.println("\nWANTED---ArrayTypeDeclarator Found in primaryExpression\n");}
    |   SUPER{addToWordsCount(1);addToWordsLength(5);}
	{updateOPCounter("super");}
    ;
*/
    
explicitConstructorCall
    :   ^(THIS_CONSTRUCTOR_CALL genericTypeArgumentList? arguments) //{System.out.println(" this constructor Call: "+$explicitConstructorCall.text);}// To recognize this();
    |   ^(SUPER_CONSTRUCTOR_CALL primaryExpression? genericTypeArgumentList? arguments)//{System.out.println(" super constructor is Called: "+$explicitConstructorCall.text);}// To recognize super()
    ;

arrayTypeDeclarator
    :   ^(ARRAY_DECLARATOR (arrayTypeDeclarator | qualifiedIdentifier | primitiveType))
    ;

newExpression
    :   ^(  STATIC_ARRAY_CREATOR
            (   primitiveType newArrayConstruction {System.out.println("\nWANTED---newExpression primitiveType newArrayConstruction\n");}
            |   genericTypeArgumentList? qualifiedTypeIdent newArrayConstruction {System.out.println("\nWANTED---newExpression  genericTypeArgumentList? qualifiedTypeIdent newArrayConstruction\n");}
            )
        )
    |   ^(CLASS_CONSTRUCTOR_CALL genericTypeArgumentList? qualifiedTypeIdent arguments classTopLevelScope?)//{System.out.println("---New Cons. call: " + $qualifiedTypeIdent.text+" arg: "+$arguments.text);}
    ;

innerNewExpression // something like 'InnerType innerType = outer.new InnerType();'
    :   ^(CLASS_CONSTRUCTOR_CALL genericTypeArgumentList? IDENT arguments classTopLevelScope?)
    ;
    
newArrayConstruction
    :   arrayDeclaratorList arrayInitializer
    |   expression+ arrayDeclaratorList?
    ;

arguments
    :   ^(ARGUMENT_LIST expressionList*)
    ;

literal 
    :   HEX_LITERAL
    |   OCTAL_LITERAL
    |   DECIMAL_LITERAL //{System.out.println("Literal Text: "+$text); const_count++;}
    |   FLOATING_POINT_LITERAL//{System.out.println("Literal Text: "+$text); const_count++;}
    |   CHARACTER_LITERAL//{System.out.println("Literal Text: "+$text); const_count++;}
    |   STRING_LITERAL //TODO strings might be complex e.g. in case of System.out.println("DotThis.f()");
    |   TRUE
    |   FALSE
    |   NULL
    ;
